import json
import logging
import os
from datetime import datetime

from microwler.settings import Settings

logging.basicConfig(level=logging.INFO, format='[%(levelname)s] %(asctime)s %(message)s', datefmt='%Y-%m-%d %H:%M:%S')


class BaseExporter:
    """
    Use this class to build your custom export functionality, i.e. send data per HTTP or SMTP.
    The Crawler instance will call export() once it's done with everything else.
    You can plug your Exporter into the crawler by adding the class to settings['exporters']
    """

    def __init__(self, domain: str, data: list, settings: Settings):
        self.domain = domain
        self.data = [page.__dict__ for page in data]
        self.settings = settings

    def export(self):
        raise NotImplementedError()


class FileExporter(BaseExporter):
    extension = ''

    def convert(self):
        raise NotImplementedError()

    def export(self):
        data = self.convert()
        timestamp = datetime.now().strftime('%Y-%m-%d-%H:%M')
        path = os.path.join(self.settings.export_to, f'{self.domain}_{timestamp}.{self.extension}')
        try:
            os.makedirs(self.settings.export_to, exist_ok=True)
            with open(path, 'w') as file:
                file.write(data)
            logging.info(f'Exported data as {self.extension.upper()} to: {path}')
        except Exception as e:
            logging.error(f'Error during export: {e}')


class JSONExporter(FileExporter):
    extension = 'json'

    def convert(self):
        data = json.dumps(self.data)
        return data


class CSVExporter(FileExporter):
    extension = 'csv'

    def convert(self):
        headers = ';'.join([key.upper() for key in self.data[0].keys()])
        rows = '\n'.join([';'.join([str(val) for val in obj.values()]) for obj in self.data])
        table = '\n'.join([headers, rows])
        return table


class HTMLExporter(FileExporter):
    extension = 'html'

    def convert(self):
        styles = 'width: 100%; border: 1px solid grey; text-align: center'
        headers = ''.join([f'<th>{key.upper()}</th>' for key in self.data[0].keys()])
        rows = ''.join([f"<tr>{''.join([f'<td>{val}</td>' for val in obj.values()])}</tr>" for obj in self.data])
        table = f'<!DOCTYPE html><html><body><table style="{styles}"><tr>{headers}</tr><tbody>{rows}</tbody></table><body></html>'
        return table






import asyncio
import logging
import time
from typing import Callable

import aiohttp
from urllib.parse import urlparse, urlencode, parse_qsl

import prettytable
from lxml import html as DOMParser

from microwler.scrape import Page
from microwler.settings import Settings
from microwler import utils

logging.basicConfig(level=logging.INFO, format='[%(levelname)s] %(asctime)s %(message)s', datefmt='%Y-%m-%d %H:%M:%S')


class Crawler:
    """
    Every crawl will be executed by a single Crawler instance.
    """

    def __init__(self,
                 start_url: str,
                 selectors: dict = None,
                 transformer: Callable[[dict], dict] = None,
                 settings: dict = None):
        """
        Setup a new Crawler instance
        Arguments:
            start_url: the URL to start crawling
            selectors: A `dict` with *selectors* [(read more)](/scraping#selectors)
            transformer: Function to transform scraped data after crawling (check out [an example](/examples#advanced))
            settings: A `dict` with configuration parameters for the crawler [(read more)](#settings)
        """
        self._start_url = start_url
        parsed = urlparse(start_url)
        self._domain = parsed.netloc
        self._base_url = f'{parsed.scheme}://{self._domain}/'
        self._selectors = selectors
        self._transformer = transformer
        self._settings = Settings(settings)
        self._seen_urls = set()
        self._session = aiohttp.ClientSession()
        self._limiter = asyncio.BoundedSemaphore(self._settings.max_concurrency)
        self._verbose = False
        self._results = []

    async def _get(self, url):
        async with self._limiter:
            try:
                heads = utils.get_headers(self._settings.language)
                async with self._session.get(url, timeout=15, headers=heads) as response:
                    html = await response.text(encoding='utf-8')
                    if self._verbose:
                        logging.info(f'Processed: {url} [{response.status}]')
                    return html, response.status
            except TimeoutError:
                logging.warning(f'Timeout error: {url}')
                return None

    def _find_links(self, html):
        """ Extract relevant links from an HTML document """
        dom = DOMParser.fromstring(html)
        dom.make_links_absolute(self._base_url)
        links = {
            # use set to ignore local duplicates
            link for link in dom.xpath('//a/@href')
            if link.startswith(self._base_url)  # stay on this website
            if link not in self._seen_urls  # try to filter global duplicates in order to avoid extra loop steps later
            and not any([link.lower().endswith(e) for e in utils.IGNORED_EXTENSIONS])  # ignore file extensions
        }
        return list(links)

    async def _get_one(self, url):
        try:
            html, status = await self._get(url)
            links = []
            if html:
                # find internal links
                links.extend(self._find_links(html))
            return url, status, html, links
        except Exception as e:
            logging.error(f'Processing error: {e} [{url}]')
            return url, e, None

    async def _get_batch(self, to_fetch):
        futures, results = [], []
        for url in to_fetch:
            normalized_url = utils.norm_url(url)
            if normalized_url in self._seen_urls:
                continue
            self._seen_urls.add(normalized_url)
            futures.append(self._get_one(normalized_url))

        for future in asyncio.as_completed(futures):
            try:
                results.append((await future))
            except Exception as e:
                logging.warning(f'Encountered exception: {e}')
        return results

    async def _crawl(self) -> [Page]:
        pipeline = [self._start_url]
        pages = []
        try:
            for depth in range(self._settings.max_depth + 1):
                batch = await self._get_batch(pipeline)
                pipeline = []
                for url, status, html, links in batch:
                    # If links is None, there was an error
                    if links is not None:
                        # Queue the URLs found on this page
                        pipeline.extend(links)
                    # Append result object
                    obj = Page(url, status, depth, list(links), html)
                    pages.append(obj)
        finally:
            await self._session.close()
            return pages

    def run(self, verbose: bool = False, sort_urls: bool = False, keep_source: bool = False):
        """
        Starts the crawler instance. You can retrieve the results using the `crawler.data` or `crawler.pages` properties.
        Arguments:
            verbose: log progress to `stdout` while crawling
            sort_urls: sort result list by URL
            keep_source: per default, the page content will be after scraping
        """
        self._verbose = verbose
        self._results = None
        start = time.time()
        future = asyncio.Task(self._crawl())
        loop = asyncio.get_event_loop()
        logging.info(f'Crawler started [{self._start_url}]')
        loop.run_until_complete(future)
        loop.close()
        logging.info(f'Crawler stopped [{self._start_url}]')
        pages = future.result()
        duration = time.time() - start

        if len(pages):
            # SORT #
            if sort_urls:
                logging.info(f'Sorting results ...')
                pages.sort(key=lambda item: item.url)

            # SELECT #
            if self._selectors:
                logging.info(f'Scraping data ...')
                pages = [page.scrape(self._selectors, keep_source=keep_source) for page in pages]

                # TRANSFORM #
                if self._transformer is not None:
                    logging.info(f'Applying transformer ...')
                    pages = [page.transform(self._transformer) for page in pages]

            # EXPORT #
            if len(self._settings.exporters):
                logging.info('Running exporters ...')
                for exporter_cls in self._settings.exporters:
                    instance = exporter_cls(self._domain, pages, self._settings)
                    instance.export()

        self._results = pages

        table = prettytable.PrettyTable()
        table.add_column('Pages', [len(self._results)])
        table.add_column('Duration', [f'{round(duration, 2)}s'])
        table.add_column('Average', [f'{round(len(self._results) / duration, 2)} p/s'])
        print(table)

    @property
    def data(self) -> [dict]:
        return [{'url': page.url, 'data': page.data if self._selectors else page.html} for page in self._results]

    @property
    def pages(self) -> [Page]:
        return self._results



import os


class Settings(object):
    max_depth: int = 10
    max_concurrency: int = 20
    language: str = 'en-us'
    export_to = os.path.join(os.getcwd(), 'exports')
    exporters: list = []

    def __init__(self, params: dict):
        if params:
            for key, value in params.items():
                setattr(self, key, value)



from setuptools import setup
import pathlib

here = pathlib.Path(__file__).parent.resolve()

long_description = (here / 'README.md').read_text(encoding='utf-8')

setup(
    name='microwler',
    version='0.1.1',
    description='A micro-framework for asynchronous deep crawls and web scraping written in Python',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://innovinati.github.io/microwler/',
    author='Maximilian Wolf',
    author_email='maximilian.wolf@innovinati.com',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Framework :: AsyncIO',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3 :: Only',
    ],
    packages=['microwler'],
    python_requires='>=3.6, <4',
    install_requires=['aiohttp', 'lxml', 'prettytable', 'random-user-agent', 'html-text'],
    project_urls={
        'Bug Reports': 'https://github.com/INNOVINATI/microwler/issues',
        'Source': 'https://github.com/INNOVINATI/microwler/',
    },
)



import logging

from html_text import extract_text
from lxml.etree import ParserError
from lxml import html as DOMParser

from microwler.utils import get_first_or_list


class Page(object):
    """
    Internal representation of a webpage

    > The methods `scrape()` and `transform()` will be called by the crawler instance
    > if selectors and a corresponding transformer function are defined.
    """

    def __init__(self, url: str, status_code: int, depth: int, links: list = None, html: bytes = None):
        """
        Arguments:
            url: the URL of this page
            status_code: the request's status code
            depth: the depth at which this page was crawled
            links: list of internal links found on this page
            html: HTML body
        """
        self.url = url
        self.status_code = status_code
        self.depth = depth
        self.links = links
        self.html = html
        self.data = {}

    def scrape(self, selectors: dict, keep_source=False):
        """
        Extracts data using the given selectors. Selectors are either `XPaths` (strings) or callables.
        If a callable is given, it will receive the parsed DOM as only argument, which is an [lxml.html.HtmlElement](https://lxml.de/api/lxml.html.HtmlElement-class.html) instance.
        This means you can apply `dom.xpath(...)` or `dom.css(...)` from `lxml.etree._Element` and return whatever you want.

        > Note: The selected data items will be stored in self.data
        """

        try:
            dom = DOMParser.fromstring(self.html)
            for field, selector in selectors.items():
                if type(selector) == str:
                    self.data[field] = get_first_or_list(dom.xpath(selector))
                else:
                    self.data[field] = get_first_or_list(selector(dom))
        except ParserError as e:
            logging.warning(f'Parsing error: {e}')

        if not keep_source:
            del self.html

        return self

    def transform(self, func):
        """
        Applies a given transformer function to this page's data.

        Arguments:
            func: a Python callable
        """
        if len(self.data):
            try:
                self.data = func(self.data)
                return self
            except Exception as e:
                logging.warning(f'Transformer error: {e}')
                return self
        raise ValueError('You need to provide selectors in order to use a transformer')


def title(dom):
    """ Extract <title> tag """
    return dom.xpath('//title//text()')


def headings(dom):
    """ Extract heading tags, i.e. <h1>, <h2>, ... """
    return {
        'h1': dom.xpath('//h1//text()'),
        'h2': dom.xpath('//h2//text()'),
        'h3': dom.xpath('//h3//text()'),
        'h4': dom.xpath('//h4//text()'),
        'h5': dom.xpath('//h5//text()'),
        'h6': dom.xpath('//h6//text()')
    }


def paragraphs(dom):
    """ Extract <p> tags """
    return dom.xpath('//p//text()')


def text(dom):
    """ Extract and clean text content """
    return extract_text(str(dom))


def meta(dom):
    """ Extract <meta> tags """
    tags = dom.xpath('//meta')
    return {tag.get('name'): tag.get('content') for tag in tags}


def canonicals(dom):
    """ Extract <link rel='canonical'> tags """
    return dom.xpath('//link[@rel=’canonical’]/@href')


def schemas(dom):
    """ Extract itemtype schemas """
    schema_links = dom.xpath('//*[@itemtype]/@itemtype')
    return [link.split('/')[-1] for link in schema_links]


def emails(dom):
    """ Extract email addresses from <a> tags """
    hrefs = dom.xpath('//a[starts-with(@href, "mailto")]/@href')
    return [href.strip('mailto:') for href in hrefs]


def images(dom):
    """ Extract URLs from <img> tags """
    return dom.xpath('//img/@src')
